import random
from torch.utils.data import Dataset, DataLoader

from .iml_datasets import JsonDataset, ManiDataset

from ..transforms import get_albu_transforms
from .utils import pil_loader, denormalize

from IMDLBenCo.registry import DATASETS
@DATASETS.register_module()
class BalancedDataset(Dataset):
    """The BalancedDataset manages multiple iml_datasets, so it does not inherit from AbstractDataset.

    Args:
        Dataset (_type_): _description_

    Returns:
        _type_: _description_
    """

    def __init__(self, 
                 path = None, 
                 sample_number = 1840,
                 *args, 
                 **kwargs
                ) -> None:
        self.sample_number = sample_number
        if path == None:
            # Defalut
            self.settings_list = [
                ['/mnt/data0/public_datasets/IML/CASIA2.0', ManiDataset],
                ['/mnt/data0/public_datasets/IML/FantasticReality_v1/FantasticReality.json', JsonDataset],
                ['/mnt/data0/public_datasets/IML/IMD_20_1024', ManiDataset],
                ['/mnt/data0/public_datasets/IML/tampCOCO/sp_COCO_list.json', JsonDataset],
                ['/mnt/data0/public_datasets/IML/tampCOCO/cm_COCO_list.json', JsonDataset],
                ['/mnt/data0/public_datasets/IML/tampCOCO/bcm_COCO_list.json', JsonDataset],
                ['/mnt/data0/public_datasets/IML/tampCOCO/bcmc_COCO_list.json', JsonDataset]
            ]
        else:
            import json
            with open(path, "r") as f:
                setting_json = json.load(f)
            # self.settings_list = path_list
            self.settings_list = []
            for dataset_str,  dataset_path in setting_json:
                self.settings_list.append(
                    [
                        dataset_path,
                        DATASETS.get(dataset_str)
                    ]
                )
            

        self.dataset_list = [self._get_dataset(path, dataset_type, *args, **kwargs) for path, dataset_type in self.settings_list]
        
        
    def _get_dataset(self, path, dataset_type, *args, **kwargs):
        return dataset_type(path, *args, **kwargs)
        
    def __len__(self):
        return self.sample_number * len(self.settings_list)    
    
    def __getitem__(self, index):
        dataset_index = index // self.sample_number

        selected_dataset = self.dataset_list[dataset_index]
        length = len(selected_dataset)
        selected_item = random.randint(0, length - 1)
        return selected_dataset[selected_item]

    def __str__(self):
        # Print basic information about the dataset and its settings
        info = f"<===BalancedDataset with {len(self.settings_list)} datasets:===>\n"
        
        total_sample = 0
        # Iterate through the settings list and print dataset paths and types
        for idx, (i_dataset) in enumerate(self.dataset_list):
            i_len = len(i_dataset)
            total_sample += i_len
            info += f"  Dataset {idx + 1}: {i_dataset}\n"
        
        # Print sample number and total number of samples
        info += f"Sample number per dataset: {self.sample_number:,}\n"
        epoch_samples = self.__len__()
        info += f"Samples per Epoch: {epoch_samples:,}\n"
        info += f"Total images in whole dataset: {total_sample:,}\n"
        info += f"<================================================>\n"
        return info



import os
import argparse
import requests

from colorama import init, Fore, Style

# from IMDLBenCo.utils.paths import BencoPath
from IMDLBenCo.cli_funcs import cli_init, cli_guide, cli_data
import importlib.metadata 

PYPI_API_URL = 'https://pypi.org/pypi/IMDLBenCo/json'
from IMDLBenCo.version import __version__

def version_and_check_for_updates():
    print(f'IMDLBenCo codebase version: {__version__}')
    try:
        response = requests.get(PYPI_API_URL, timeout=5)
        response.raise_for_status()  # 如果响应码不是200，则抛出异常
        pypi_data = response.json()
        cloud_version = pypi_data['info']['version']  # 获取最新版本号
        # cloud_version = '0.1.21'   # for debug & test
        print("\tChecking for updates...")
        print("\tLocal version: ", __version__)
        print("\tPyPI newest version: ", cloud_version)
        local_version = __version__  # 通过 importlib.metadata 获取当前安装版本

        if cloud_version != local_version:
            print(Fore.YELLOW + f"New version available: {cloud_version}. Your version: {local_version}.")
            print("Run 'pip install --upgrade imdlbenco' to upgrade.")           
        else:
            print(Fore.GREEN + f"You are using the latest version: {local_version}.")
    except requests.exceptions.RequestException as e:
        print(Fore.RED + "Failed to check for updates from PyPI. Please check your internet connection.")
        print(f"Error: {e}")

def main():
    parser = argparse.ArgumentParser(description='Command line interface for IMDLBenCo, with codebase version: ' + __version__)
    
    # Add version argument with update check only when user requests version
    parser.add_argument('-v', '--version', action='store_true', help="Show the version of the tool")
    
    subparsers = parser.add_subparsers(dest='command', required=False)
    
    # init command
    parser_init = subparsers.add_parser('init', help='Initialize the environment')
    init_subparsers = parser_init.add_subparsers(dest='subcommand', required=False)
    
    # init base
    parser_init_base = init_subparsers.add_parser('base', help='Initialize the base environment')
    parser_init_base.set_defaults(subcommand='base')

    # init model_zoo
    parser_init_model_zoo = init_subparsers.add_parser('model_zoo', help='Initialize the model zoo')
    
    # init backbone
    parser_init_backbone = init_subparsers.add_parser('backbone', help='Initialize the backbone')
    
    
    # guide command
    parser_guide = subparsers.add_parser('guide', help='Guide for using the tool')
    
    # data command
    parser_data = subparsers.add_parser('data', help='Manage data')
    
    parser.add_argument('--config', type=str, help='Path to the configuration file')
    
    args = parser.parse_args()
    if args.version:
        version_and_check_for_updates()
    
    if args.command == 'init':
        if args.subcommand is None:
            args.subcommand = 'base'
        cli_init(args.config, subcommand=args.subcommand)
        
    elif args.command == 'guide':
        cli_guide(args.config)
    elif args.command == 'data':
        cli_data(args.config)

def train(config):
    print(f'Training with config: {config}')

def evaluate(config):
    print(f'Evaluating with config: {config}')

if __name__ == '__main__':
    main()



from cv2 import detail_ImageFeatures
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from collections import OrderedDict
#Pytorch
import torch
from torch import nn
import torch.nn.functional as F


class SRMConv2D(nn.Module):
    def __init__(self):
        super(SRMConv2D,self).__init__()
        q = [4, 12, 2] # coefficient of the kernels
        kernel1 = np.array([
            [0, 0, 0, 0, 0],
            [0,-1, 2,-1, 0],
            [0, 2,-4, 2, 0],
            [0,-1, 2,-1, 0],
            [0, 0, 0, 0, 0]
        ],dtype=np.float32)
        kernel2 = np.array([
            [-1, 2,-2, 2,-1],
            [2, -6, 8,-6, 2],
            [-2, 8,-12,8,-2],
            [2, -6, 8,-6, 2],
            [-1, 2,-2, 2,-1]          
        ],dtype=np.float32)
        kernel3 = np.array([
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 1,-2, 1, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ],dtype=np.float32)
        zero_kernel = np.zeros_like(kernel3)
        # shape (3,9,5,5)
        weight = torch.tensor(np.array([ 
            [[kernel1 / q[0] if j == i else zero_kernel for j in range(3)] for i in range(3)],
            [[kernel2 / q[1] if j == i else zero_kernel for j in range(3)] for i in range(3)],
            [[kernel3 / q[2] if j == i else zero_kernel for j in range(3)] for i in range(3)],
        ]),dtype=torch.float32)
        weight = weight.reshape(-1,3,5,5)
        self.weight = torch.nn.Parameter(weight, requires_grad=False) 

    def forward(self, x):
        with torch.no_grad():
            return torch.nn.functional.conv2d(x, weight=self.weight, padding = 2)





import os
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from torch.utils.data import DataLoader, DistributedSampler, Dataset
from collections import Counter

""" Expected output:
------------------------------
Sampler drop_last=True, DataLoader drop_last=True,
 Sampler shuffle=True, DataLoader shuffle=default
Missing data: Counter({44: 1, 52: 1, 87: 1, 100: 1})
Extra data: Counter()
------------------------------
------------------------------
Sampler drop_last=True, DataLoader drop_last=True,
 Sampler shuffle=False, DataLoader shuffle=default
Missing data: Counter({97: 1, 98: 1, 99: 1, 100: 1})
Extra data: Counter()
------------------------------
------------------------------
Sampler drop_last=True, DataLoader drop_last=False,
 Sampler shuffle=True, DataLoader shuffle=default
Missing data: Counter({52: 1})
Extra data: Counter()
------------------------------
------------------------------
Sampler drop_last=True, DataLoader drop_last=False,
 Sampler shuffle=False, DataLoader shuffle=default
Missing data: Counter({100: 1})
Extra data: Counter()
------------------------------
------------------------------
Sampler drop_last=False, DataLoader drop_last=True,
 Sampler shuffle=True, DataLoader shuffle=default
Missing data: Counter()
Extra data: Counter({45: 1, 20: 1})
------------------------------
------------------------------
Sampler drop_last=False, DataLoader drop_last=True,
 Sampler shuffle=False, DataLoader shuffle=default
Missing data: Counter()
Extra data: Counter({1: 1, 2: 1})
------------------------------
------------------------------
Sampler drop_last=False, DataLoader drop_last=False,
 Sampler shuffle=True, DataLoader shuffle=default
Missing data: Counter()
Extra data: Counter({45: 1, 20: 1})
------------------------------
------------------------------
Sampler drop_last=False, DataLoader drop_last=False,
 Sampler shuffle=False, DataLoader shuffle=default
Missing data: Counter()
Extra data: Counter({1: 1, 2: 1})
------------------------------
"""

class CustomDataset(Dataset):
    def __init__(self, data):
        self.data = data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]

def setup(rank, world_size):
    # 初始化分布式环境 Initialize distributed environment
    dist.init_process_group(backend='nccl', init_method='env://', rank=rank, world_size=world_size)

def cleanup():
    # 销毁分布式环境 Destroy the distributed environment
    dist.destroy_process_group()

def run_config(rank, world_size, sampler_drop_last, dataloader_drop_last, sampler_shuffle):
    # 设置当前进程的GPU Set the GPU for the current process
    torch.cuda.set_device(rank)
    setup(rank, world_size)

    # 创建一个简单的数据集 Create a simple dataset
    data = torch.arange(1, 101).view(-1, 1)  # 数据集包含 100 个样本 Dataset contains 100 samples
    dataset = CustomDataset(data)

    # 设置批次大小，每个GPU上为2，总批次大小为6 Set batch size, 2 per GPU, total batch size is 6
    batch_size = 2

    # 创建分布式采样器 Create distributed sampler
    sampler = DistributedSampler(dataset, num_replicas=world_size, rank=rank, drop_last=sampler_drop_last, shuffle=sampler_shuffle)

    # 创建 DataLoader Create DataLoader
    dataloader = DataLoader(dataset, batch_size=batch_size, sampler=sampler, drop_last=dataloader_drop_last)

    # 用于收集每个进程的数据 Collect data from each process
    collected_data = []

    # 模拟训练循环 Simulate training loop
    for epoch in range(1):
        sampler.set_epoch(epoch)
        for batch in dataloader:
            collected_data.extend(batch)

    # 将每个进程的数据转换为tensor Convert data from each process to tensor
    collected_data_tensor = torch.tensor(collected_data, dtype=torch.int32, device=torch.device(rank))

    # 将所有进程的数据聚合到主进程 Gather data from all processes to the master process
    gathered_data = [torch.zeros_like(collected_data_tensor) for _ in range(world_size)]
    dist.all_gather(gathered_data, collected_data_tensor)

    if rank == 0:
        # 在主进程中合并并检查数据 Merge and check data in the master process
        all_data = torch.cat(gathered_data).cpu().numpy().tolist()

        # 使用Counter来统计每个数据的出现次数 Use Counter to count the occurrence of each data item
        all_data_counter = Counter(all_data)
        expected_data_counter = Counter(range(1, 101))

        missing_data = expected_data_counter - all_data_counter
        extra_data = all_data_counter - expected_data_counter
        print("---"*10)
        print(f"Sampler drop_last={sampler_drop_last}, DataLoader drop_last={dataloader_drop_last},\n Sampler shuffle={sampler_shuffle}, DataLoader shuffle=default")
        print(f"Missing data: {missing_data}")
        print(f"Extra data: {extra_data}")
        print("---"*10)

    cleanup()

def spawn_for_config(sampler_drop_last, dataloader_drop_last, sampler_shuffle):
    world_size = 3  # 总的进程数，即GPU数量 Total number of processes (GPUs)
    mp.spawn(run_config, args=(world_size, sampler_drop_last, dataloader_drop_last, sampler_shuffle), nprocs=world_size, join=True)

if __name__ == "__main__":
    os.environ['CUDA_VISIBLE_DEVICES'] = '1,2,3'  # 指定要使用的GPU Specify which GPUs to use
    os.environ['MASTER_ADDR'] = '127.0.0.1'  # 设置主节点地址 Set the master node address
    os.environ['MASTER_PORT'] = '29501'  # 设置主节点端口 Set the master node port

    configs = [
        (sampler_drop_last, dataloader_drop_last, sampler_shuffle)
        for sampler_drop_last in [True, False]
        for dataloader_drop_last in [True, False]
        for sampler_shuffle in [True, False]
    ]

    for config in configs:
        # print(f"Running config: Sampler drop_last={config[0]}, DataLoader drop_last={config[1]}, Sampler shuffle={config[2]}, DataLoader shuffle=defaults")
        spawn_for_config(*config)


from collections import abc
from typing import Any, Callable, Optional, Type, Union

def is_seq_of(seq: Any,
            expected_type: Union[Type, tuple],
            seq_type: Type = None) -> bool:
    """Check whether it is a sequence of some type.

    Args:
        seq (Sequence): The sequence to be checked.
        expected_type (type or tuple): Expected type of sequence items.
        seq_type (type, optional): Expected sequence type. Defaults to None.

    Returns:
        bool: Return True if ``seq`` is valid else False.

    Examples:
        >>> from mmengine.utils import is_seq_of
        >>> seq = ['a', 'b', 'c']
        >>> is_seq_of(seq, str)
        True
        >>> is_seq_of(seq, int)
        False
    """
    if seq_type is None:
        exp_seq_type = abc.Sequence
    else:
        assert isinstance(seq_type, type)
        exp_seq_type = seq_type
    if not isinstance(seq, exp_seq_type):
        return False
    for item in seq:
        if not isinstance(item, expected_type):
            return False
    return True


import numpy as np
import cv2
from sklearn.metrics import f1_score
import os

# 目录路径 (Directory path)
output_dir = "output_masks_shape"

import numpy as np
import cv2
from PIL import Image
import os

# 创建目录以保存图像 (Create directory to save images)
os.makedirs(output_dir, exist_ok=True)

# 设置图像的宽度和高度 (Set the width and height of the image)
width, height = 256, 256

def draw_random_polygons(image, num_polygons=5, min_vertices=3, max_vertices=8):
    for _ in range(num_polygons):
        num_vertices = np.random.randint(min_vertices, max_vertices + 1)
        points = np.random.randint(0, min(width, height), (num_vertices, 2))
        points = points.reshape((-1, 1, 2))
        color = (np.random.randint(256),)  # 单通道颜色 (Single channel color)
        cv2.fillPoly(image, [points], color)
    return image

for i in range(1, 9):
    # 创建空的灰度图像 (Create an empty grayscale image)
    prediction_mask = np.zeros((height, width), dtype=np.uint8)
    gt_mask = np.zeros((height, width), dtype=np.uint8)

    # 在图像上绘制随机多边形 (Draw random polygons on the image)
    prediction_mask = draw_random_polygons(prediction_mask)
    gt_mask = draw_random_polygons(gt_mask) > 0.5

    # 将图像转换为Pillow图像并保存 (Convert image to Pillow image and save)
    prediction_image = Image.fromarray(prediction_mask)
    gt_image = Image.fromarray(gt_mask)

    # 保存预测mask和gt mask (Save prediction mask and ground truth mask)
    prediction_image.save(os.path.join(output_dir, f"image{i}.jpg"))
    gt_image.save(os.path.join(output_dir, f"mask{i}.jpg"))

print("图像和mask已成功生成并保存。 (Images and masks have been successfully generated and saved.)")


def calculate_binary_f1_score(pred, gt, return_cf_matrix=False):
    # 将图像展开成一维数组 (Flatten the images into 1D arrays)
    pred = pred.flatten()
    gt = gt.flatten()

    # 计算TP, FP, FN (Calculate TP, FP, FN)
    tp = np.sum((pred == 1) & (gt == 1))
    fp = np.sum((pred == 1) & (gt == 0))
    fn = np.sum((pred == 0) & (gt == 1))

    # 计算Precision和Recall (Calculate Precision and Recall)
    precision = tp / (tp + fp) if (tp + fp) != 0 else 0
    recall = tp / (tp + fn) if (tp + fn) != 0 else 0

    # 计算F1 Score (Calculate F1 Score)
    if precision + recall == 0:
        f1 = 0
    else:
        f1 = 2 * (precision * recall) / (precision + recall)
    
    if return_cf_matrix == True:
        return f1, tp, fp, fn
    else:
        return f1


def calculate_macro_f1_score(pred, gt):
    f1 = calculate_binary_f1_score(pred, gt)
    f1_hat = calculate_binary_f1_score(1-pred, 1-gt)
    return (f1 + f1_hat) / 2


def calculate_micro_f1_score(pred, gt):
    f1, tp, fp, fn = calculate_binary_f1_score(pred, gt, return_cf_matrix=True)
    f1_, tp_, fp_, fn_ = calculate_binary_f1_score(1-pred, 1-gt, return_cf_matrix=True)
    
    tp += tp_
    fp += fp_
    fn += fn_
    
    # 计算Precision和Recall (Calculate Precision and Recall)
    precision = tp / (tp + fp) if (tp + fp) != 0 else 0
    recall = tp / (tp + fn) if (tp + fn) != 0 else 0

    # 计算F1 Score (Calculate F1 Score)
    if precision + recall == 0:
        f1 = 0
    else:
        f1 = 2 * (precision * recall) / (precision + recall)
        
    return f1

# 初始化列表以存储F1 Score (Initialize list to store F1 Scores)
def check_f1(name, manual_f1):
    f1_scores = []
    manual_f1_scores = []
    for i in range(1, 9):
        # 读取预测mask和gt mask (Read prediction mask and ground truth mask)
        prediction_path = os.path.join(output_dir, f"image{i}.jpg")
        gt_path = os.path.join(output_dir, f"mask{i}.jpg")

        prediction_mask = cv2.imread(prediction_path, cv2.IMREAD_GRAYSCALE)
        gt_mask = cv2.imread(gt_path, cv2.IMREAD_GRAYSCALE)

        # 二值化图像 (Binarize the images - convert them to 0 and 1)
        _, prediction_mask_bin = cv2.threshold(prediction_mask, 127, 1, cv2.THRESH_BINARY)
        _, gt_mask_bin = cv2.threshold(gt_mask, 127, 1, cv2.THRESH_BINARY)

        # 计算F1 Score (Calculate F1 Score)
        f1 = f1_score(gt_mask_bin.flatten(), prediction_mask_bin.flatten(), average=name)
        binary_f1 = f1_score(gt_mask_bin.flatten(), prediction_mask_bin.flatten(), average="binary")
        f1_manual = manual_f1(gt_mask_bin, prediction_mask_bin)
        f1_scores.append(f1)
        manual_f1_scores.append(f1_manual)

        print(f"Image {i} {name} F1 Score: {f1}")
        print(f"Image {i} binary F1 Score: {binary_f1}")
        print(f"Image {i} {name} manual F1 score {f1_manual}")
        print(f"\033[33m{name} F1 - Binary F1 == {f1 - binary_f1}\033[0m")
        if f1 == f1_manual:
            print(f"\033[92mImage {name} F1 - {i} - same with manual implementation!\033[0m")
        else:
            print(f"\033[91mImage {name} F1 - {i} - Incorrect with manual implementation!\033[0m")
        
    # 计算平均F1 Score (Calculate average F1 Score)
    mean_f1_score = np.mean(f1_scores)
    mean_manual_f1_score = np.mean(manual_f1_scores)
    print(f"Average F1 Score: {mean_f1_score}")
    print(f"Average manual F1 Score: {mean_manual_f1_score}")
    print(40*"*", '\n')
    
    
check_f1("binary", calculate_binary_f1_score)
check_f1("macro", calculate_macro_f1_score)
check_f1("micro", calculate_micro_f1_score)



from .version import __version__, version_info
from .registry import Registry, MODELS, DATASETS, POSTFUNCS

from .datasets import * # Registry all Datasets to the DATASETS register
from .model_zoo import *
from .modules import *  

__all__ = ['__version__', 'version_info', 'MODELS', "DATASETS", "POSTFUNCS"]






import torch
import torch.nn as nn
import torch.nn.functional as F
from functools import partial

from timm.models.layers import DropPath, to_2tuple, trunc_normal_
from ..net_utils import FeatureFusionModule as FFM
from ..net_utils import FeatureRectifyModule as FRM
import math
import time
#from engine.logger import get_logger
import logging as logger

#logger = get_logger()


class DWConv(nn.Module):
    """
    Depthwise convolution bloc: input: x with size(B N C); output size (B N C)
    """
    def __init__(self, dim=768):
        super(DWConv, self).__init__()
        self.dwconv = nn.Conv2d(dim, dim, kernel_size=3, stride=1, padding=1, bias=True, groups=dim)

    def forward(self, x, H, W):
        B, N, C = x.shape
        x = x.permute(0, 2, 1).reshape(B, C, H, W).contiguous() # B N C -> B C N -> B C H W
        x = self.dwconv(x) 
        x = x.flatten(2).transpose(1, 2) # B C H W -> B N C

        return x


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        """
        MLP Block: 
        """
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = self.fc1(x)
        x = self.dwconv(x, H, W)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0., sr_ratio=1):
        super().__init__()
        assert dim % num_heads == 0, f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        # Linear embedding
        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        B, N, C = x.shape
        # B N C -> B N num_head C//num_head -> B C//num_head N num_heads
        q = self.q(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3) 

        if self.sr_ratio > 1:
            x_ = x.permute(0, 2, 1).reshape(B, C, H, W) 
            x_ = self.sr(x_).reshape(B, C, -1).permute(0, 2, 1) 
            x_ = self.norm(x_)
            kv = self.kv(x_).reshape(B, -1, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4) 
        else:
            kv = self.kv(x).reshape(B, -1, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4) 
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class Block(nn.Module):
    """
    Transformer Block: Self-Attention -> Mix FFN -> OverLap Patch Merging
    """
    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, sr_ratio=1):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale,
            attn_drop=attn_drop, proj_drop=drop, sr_ratio=sr_ratio)
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))

        return x


class OverlapPatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """

    def __init__(self, img_size=224, patch_size=7, stride=4, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)

        self.img_size = img_size
        self.patch_size = patch_size
        self.H, self.W = img_size[0] // patch_size[0], img_size[1] // patch_size[1]
        self.num_patches = self.H * self.W
        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=stride,
                              padding=(patch_size[0] // 2, patch_size[1] // 2))
        self.norm = nn.LayerNorm(embed_dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x):
        # B C H W
        x = self.proj(x)
        _, _, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        # B H*W/16 C
        x = self.norm(x)

        return x, H, W


class RGBXTransformer(nn.Module):
    def __init__(self, img_size=224, patch_size=16, in_chans=3, num_classes=1000, embed_dims=[64, 128, 256, 512], 
                 num_heads=[1, 2, 4, 8], mlp_ratios=[4, 4, 4, 4], qkv_bias=False, qk_scale=None, drop_rate=0.,
                 attn_drop_rate=0., drop_path_rate=0., norm_layer=nn.LayerNorm, norm_fuse=nn.BatchNorm2d,
                 depths=[3, 4, 6, 3], sr_ratios=[8, 4, 2, 1], stride0=4):
        super().__init__()
        self.num_classes = num_classes
        self.depths = depths

        # patch_embed
        self.patch_embed1 = OverlapPatchEmbed(img_size=img_size, patch_size=7, stride=stride0, in_chans=in_chans,
                                              embed_dim=embed_dims[0])
        self.patch_embed2 = OverlapPatchEmbed(img_size=img_size // 4, patch_size=3, stride=2, in_chans=embed_dims[0],
                                              embed_dim=embed_dims[1])
        self.patch_embed3 = OverlapPatchEmbed(img_size=img_size // 8, patch_size=3, stride=2, in_chans=embed_dims[1],
                                              embed_dim=embed_dims[2])
        self.patch_embed4 = OverlapPatchEmbed(img_size=img_size // 16, patch_size=3, stride=2, in_chans=embed_dims[2],
                                              embed_dim=embed_dims[3])

        self.extra_patch_embed1 = OverlapPatchEmbed(img_size=img_size, patch_size=7, stride=stride0, in_chans=in_chans,
                                              embed_dim=embed_dims[0])
        self.extra_patch_embed2 = OverlapPatchEmbed(img_size=img_size // 4, patch_size=3, stride=2, in_chans=embed_dims[0],
                                              embed_dim=embed_dims[1])
        self.extra_patch_embed3 = OverlapPatchEmbed(img_size=img_size // 8, patch_size=3, stride=2, in_chans=embed_dims[1],
                                              embed_dim=embed_dims[2])
        self.extra_patch_embed4 = OverlapPatchEmbed(img_size=img_size // 16, patch_size=3, stride=2, in_chans=embed_dims[2],
                                              embed_dim=embed_dims[3])

        # transformer encoder
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule
        cur = 0

        self.block1 = nn.ModuleList([Block(
            dim=embed_dims[0], num_heads=num_heads[0], mlp_ratio=mlp_ratios[0], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[0])
            for i in range(depths[0])])
        self.norm1 = norm_layer(embed_dims[0])

        self.extra_block1 = nn.ModuleList([Block(
            dim=embed_dims[0], num_heads=num_heads[0], mlp_ratio=mlp_ratios[0], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[0])
            for i in range(depths[0])])
        self.extra_norm1 = norm_layer(embed_dims[0])
        cur += depths[0]

        self.block2 = nn.ModuleList([Block(
            dim=embed_dims[1], num_heads=num_heads[1], mlp_ratio=mlp_ratios[1], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur], norm_layer=norm_layer,
            sr_ratio=sr_ratios[1])
            for i in range(depths[1])])
        self.norm2 = norm_layer(embed_dims[1])

        self.extra_block2 = nn.ModuleList([Block(
            dim=embed_dims[1], num_heads=num_heads[1], mlp_ratio=mlp_ratios[1], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur+1], norm_layer=norm_layer,
            sr_ratio=sr_ratios[1])
            for i in range(depths[1])])
        self.extra_norm2 = norm_layer(embed_dims[1])

        cur += depths[1]

        self.block3 = nn.ModuleList([Block(
            dim=embed_dims[2], num_heads=num_heads[2], mlp_ratio=mlp_ratios[2], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[2])
            for i in range(depths[2])])
        self.norm3 = norm_layer(embed_dims[2])

        self.extra_block3 = nn.ModuleList([Block(
            dim=embed_dims[2], num_heads=num_heads[2], mlp_ratio=mlp_ratios[2], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[2])
            for i in range(depths[2])])
        self.extra_norm3 = norm_layer(embed_dims[2])

        cur += depths[2]

        self.block4 = nn.ModuleList([Block(
            dim=embed_dims[3], num_heads=num_heads[3], mlp_ratio=mlp_ratios[3], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[3])
            for i in range(depths[3])])
        self.norm4 = norm_layer(embed_dims[3])

        self.extra_block4 = nn.ModuleList([Block(
            dim=embed_dims[3], num_heads=num_heads[3], mlp_ratio=mlp_ratios[3], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[3])
            for i in range(depths[3])])
        self.extra_norm4 = norm_layer(embed_dims[3])

        cur += depths[3]

        self.FRMs = nn.ModuleList([
                    FRM(dim=embed_dims[0], reduction=1),
                    FRM(dim=embed_dims[1], reduction=1),
                    FRM(dim=embed_dims[2], reduction=1),
                    FRM(dim=embed_dims[3], reduction=1)])

        self.FFMs = nn.ModuleList([
                    FFM(dim=embed_dims[0], reduction=1, num_heads=num_heads[0], norm_layer=norm_fuse),
                    FFM(dim=embed_dims[1], reduction=1, num_heads=num_heads[1], norm_layer=norm_fuse),
                    FFM(dim=embed_dims[2], reduction=1, num_heads=num_heads[2], norm_layer=norm_fuse),
                    FFM(dim=embed_dims[3], reduction=1, num_heads=num_heads[3], norm_layer=norm_fuse)])

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def init_weights(self, pretrained=None):
        if isinstance(pretrained, str):
            load_dualpath_model(self, pretrained)
        else:
            raise TypeError('pretrained must be a str or None')
        
    def forward_features(self, x_rgb, x_e):
        """
        x_rgb: B x N x H x W
        """
        B = x_rgb.shape[0]
        outs = []
        outs_fused = []

        # stage 1
        x_rgb, H, W = self.patch_embed1(x_rgb)
        # B H*W/16 C
        x_e, _, _ = self.extra_patch_embed1(x_e)
        for i, blk in enumerate(self.block1):
            x_rgb = blk(x_rgb, H, W)
        for i, blk in enumerate(self.extra_block1):
            x_e = blk(x_e, H, W)
        x_rgb = self.norm1(x_rgb)
        x_e = self.extra_norm1(x_e)

        x_rgb = x_rgb.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        x_e = x_e.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        x_rgb, x_e = self.FRMs[0](x_rgb, x_e)
        x_fused = self.FFMs[0](x_rgb, x_e)
        outs.append(x_fused)
        

        # stage 2
        x_rgb, H, W = self.patch_embed2(x_rgb)
        x_e, _, _ = self.extra_patch_embed2(x_e)
        for i, blk in enumerate(self.block2):
            x_rgb = blk(x_rgb, H, W)
        for i, blk in enumerate(self.extra_block2):
            x_e = blk(x_e, H, W)
        x_rgb = self.norm2(x_rgb)
        x_e = self.extra_norm2(x_e)

        x_rgb = x_rgb.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        x_e = x_e.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        x_rgb, x_e = self.FRMs[1](x_rgb, x_e)
        x_fused = self.FFMs[1](x_rgb, x_e)
        outs.append(x_fused)
        

        # stage 3
        x_rgb, H, W = self.patch_embed3(x_rgb)
        x_e, _, _ = self.extra_patch_embed3(x_e)
        for i, blk in enumerate(self.block3):
            x_rgb = blk(x_rgb, H, W)
        for i, blk in enumerate(self.extra_block3):
            x_e = blk(x_e, H, W)
        x_rgb = self.norm3(x_rgb)
        x_e = self.extra_norm3(x_e)

        x_rgb = x_rgb.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        x_e = x_e.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        x_rgb, x_e = self.FRMs[2](x_rgb, x_e)
        x_fused = self.FFMs[2](x_rgb, x_e)
        outs.append(x_fused)
        

        # stage 4
        x_rgb, H, W = self.patch_embed4(x_rgb)
        x_e, _, _ = self.extra_patch_embed4(x_e)
        for i, blk in enumerate(self.block4):
            x_rgb = blk(x_rgb, H, W)
        for i, blk in enumerate(self.extra_block4):
            x_e = blk(x_e, H, W)
        x_rgb = self.norm4(x_rgb)
        x_e = self.extra_norm4(x_e)

        x_rgb = x_rgb.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        x_e = x_e.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        x_rgb, x_e = self.FRMs[3](x_rgb, x_e)
        x_fused = self.FFMs[3](x_rgb, x_e)
        outs.append(x_fused)
        
        return outs

    def forward(self, x_rgb, x_e):
        out = self.forward_features(x_rgb, x_e)
        return out


def load_dualpath_model(model, model_file):
    # load raw state_dict
    t_start = time.time()
    if isinstance(model_file, str):
        raw_state_dict = torch.load(model_file, map_location=torch.device('cpu'))
        #raw_state_dict = torch.load(model_file)
        if 'model' in raw_state_dict.keys():
            raw_state_dict = raw_state_dict['model']
    else:
        raw_state_dict = model_file
    
    state_dict = {}
    for k, v in raw_state_dict.items():
        if k.find('patch_embed') >= 0:
            state_dict[k] = v
            state_dict[k.replace('patch_embed', 'extra_patch_embed')] = v
        elif k.find('block') >= 0:
            state_dict[k] = v
            state_dict[k.replace('block', 'extra_block')] = v
        elif k.find('norm') >= 0:
            state_dict[k] = v
            state_dict[k.replace('norm', 'extra_norm')] = v

    t_ioend = time.time()

    model.load_state_dict(state_dict, strict=False)
    del state_dict
    
    t_end = time.time()
    logger.info(
        "Load model, Time usage:\n\tIO: {}, initialize parameters: {}".format(
            t_ioend - t_start, t_end - t_ioend))


class mit_b0(RGBXTransformer):
    def __init__(self, fuse_cfg=None, stride0=4, **kwargs):
        super(mit_b0, self).__init__(
            patch_size=4, embed_dims=[32, 64, 160, 256], num_heads=[1, 2, 5, 8], mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), depths=[2, 2, 2, 2], sr_ratios=[8, 4, 2, 1],
            drop_rate=0.0, drop_path_rate=0.1, stride0=stride0)


class mit_b1(RGBXTransformer):
    def __init__(self, fuse_cfg=None, stride0=4, **kwargs):
        super(mit_b1, self).__init__(
            patch_size=4, embed_dims=[64, 128, 320, 512], num_heads=[1, 2, 5, 8], mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), depths=[2, 2, 2, 2], sr_ratios=[8, 4, 2, 1],
            drop_rate=0.0, drop_path_rate=0.1, stride0=stride0)


class mit_b2(RGBXTransformer):
    def __init__(self, fuse_cfg=None, stride0=4, **kwargs):
        super(mit_b2, self).__init__(
            patch_size=4, embed_dims=[64, 128, 320, 512], num_heads=[1, 2, 5, 8], mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), depths=[3, 4, 6, 3], sr_ratios=[8, 4, 2, 1],
            drop_rate=0.0, drop_path_rate=0.1, stride0=stride0)


class mit_b3(RGBXTransformer):
    def __init__(self, fuse_cfg=None, stride0=4, **kwargs):
        super(mit_b3, self).__init__(
            patch_size=4, embed_dims=[64, 128, 320, 512], num_heads=[1, 2, 5, 8], mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), depths=[3, 4, 18, 3], sr_ratios=[8, 4, 2, 1],
            drop_rate=0.0, drop_path_rate=0.1, stride0=stride0)


class mit_b4(RGBXTransformer):
    def __init__(self, fuse_cfg=None, stride0=4, **kwargs):
        super(mit_b4, self).__init__(
            patch_size=4, embed_dims=[64, 128, 320, 512], num_heads=[1, 2, 5, 8], mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), depths=[3, 8, 27, 3], sr_ratios=[8, 4, 2, 1],
            drop_rate=0.0, drop_path_rate=0.1, stride0=stride0)


class mit_b5(RGBXTransformer):
    def __init__(self, fuse_cfg=None, stride0=4, **kwargs):
        super(mit_b5, self).__init__(
            patch_size=4, embed_dims=[64, 128, 320, 512], num_heads=[1, 2, 5, 8], mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), depths=[3, 6, 40, 3], sr_ratios=[8, 4, 2, 1],
            drop_rate=0.0, drop_path_rate=0.1, stride0=stride0)


import os
import subprocess
import tempfile
import platform

def get_default_editor():
    """Get the system's default editor."""
    editor = os.getenv('EDITOR')
    if not editor:
        if platform.system() == 'Windows':
            editor = 'notepad'
        elif platform.system() == 'Darwin':  # macOS
            editor = 'nano'  # macOS does not have a default GUI editor
        else:  # Linux and other Unix-like systems
            editor = 'code --wait'  # 'nano' is a common text editor in Unix-like systems
    return editor

def open_file_with_editor(file_path):
    """Open the file with the default editor."""
    editor = get_default_editor()
    try:
        subprocess.run([editor, file_path])
    except Exception as e:
        print(f"Error opening file with editor {editor}: {e}")

def commit_message():
    """Simulate the Git commit message process."""
    with tempfile.NamedTemporaryFile(delete=False, mode='w+', suffix=".tmp") as temp_file:
        temp_file_path = temp_file.name
        temp_file.write("# Please enter the commit message for your changes. Lines starting\n")
        temp_file.write("# with '#' will be ignored, and an empty message aborts the commit.\n")
        temp_file.write("\n")
        temp_file.write("# This is a comment line.\n")
        temp_file.flush()
        open_file_with_editor(temp_file_path)
        
        with open(temp_file_path, 'r') as f:
            commit_msg = ''.join([line for line in f if not line.startswith('#')]).strip()

        os.remove(temp_file_path)
        
        if not commit_msg:
            print("Aborting commit due to empty commit message.")
        else:
            print(f"Commit message is:\n{commit_msg}")

if __name__ == "__main__":
    commit_message()



import torch
from torch.utils.data import DataLoader, Dataset

# 创建一个简单的自定义Dataset
class MyDataset(Dataset):
    def __init__(self):
        self.data = [1, 2, 3, 4, 5]
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        return self.data[idx]

# 实例化Dataset
dataset = MyDataset()

# 用DataLoader加载Dataset
dataloader = DataLoader(dataset, batch_size=2, shuffle=True)

print(len(dataloader.dataset))
# 获取DataLoader的Dataset引用
dataset_from_dataloader = dataloader.dataset

# 验证引用是否相同
print(f"Original Dataset ID: {id(dataset)}")
print(f"Dataset from DataLoader ID: {id(dataset_from_dataloader)}")
print(f"Is same object: {dataset is dataset_from_dataloader}")


import torch
import torch.nn as nn
import torchvision.transforms as transforms
import matplotlib.pyplot as plt
import numpy as np
import inspect
class SobelFilter(nn.Module):
    def __init__(self,in_chan=3, out_chan=1,norm = None ):
        super(SobelFilter, self).__init__()
        if norm is not None and (not inspect.isclass(norm) or not issubclass(norm, nn.Module)):
            raise ValueError("norm must be a class derived from nn.Module or None")
        
        filter_x = np.array([
            [1, 0, -1],
            [2, 0, -2],
            [1, 0, -1],
        ]).astype(np.float32)
        filter_y = np.array([
            [1, 2, 1],
            [0, 0, 0],
            [-1, -2, -1],
        ]).astype(np.float32)

        filter_x = filter_x.reshape((1, 1, 3, 3))
        filter_x = np.repeat(filter_x, in_chan, axis=1)
        filter_x = np.repeat(filter_x, out_chan, axis=0)

        filter_y = filter_y.reshape((1, 1, 3, 3))
        filter_y = np.repeat(filter_y, in_chan, axis=1)
        filter_y = np.repeat(filter_y, out_chan, axis=0)

        filter_x = torch.from_numpy(filter_x)
        filter_y = torch.from_numpy(filter_y)
        filter_x = nn.Parameter(filter_x, requires_grad=False)
        filter_y = nn.Parameter(filter_y, requires_grad=False)

        sobel_kernel_x = torch.tensor([[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]], dtype=torch.float32).unsqueeze(0).unsqueeze(0)
        sobel_kernel_y = torch.tensor([[-1, -2, -1], [0, 0, 0], [1, 2, 1]], dtype=torch.float32).unsqueeze(0).unsqueeze(0)
        conv_x = nn.Conv2d(in_chan, out_chan, kernel_size=3, stride=1, padding=1, bias=False)
        conv_x.weight = filter_x
        conv_y = nn.Conv2d(in_chan, out_chan, kernel_size=3, stride=1, padding=1, bias=False)
        conv_y.weight = filter_y
        norm1 = nn.Identity() if norm is None else norm(out_chan)
        norm2 = nn.Identity() if norm is None else norm(out_chan)

        self.sobel_x = nn.Sequential(conv_x, norm1)
        self.sobel_y = nn.Sequential(conv_y, norm2)


        self.sobel_kernel_x = sobel_kernel_x.expand(1, 3, 3, 3)
        self.sobel_kernel_y = sobel_kernel_y.expand(1, 3, 3, 3)

        self.conv_x = nn.Conv2d(in_channels=3, out_channels=1, kernel_size=3, padding=1, bias=False)
        self.conv_y = nn.Conv2d(in_channels=3, out_channels=1, kernel_size=3, padding=1, bias=False)

        
        self.conv_x.weight = nn.Parameter(self.sobel_kernel_x, requires_grad=False)
        self.conv_y.weight = nn.Parameter(self.sobel_kernel_y, requires_grad=False)

    def forward(self, x):
        grad_x = self.conv_x(x)
        grad_y = self.conv_y(x)
        grad = torch.sqrt(grad_x**2 + grad_y**2)
        return grad




from .windowed_attention_vit import ViT as window_attention_vit, SimpleFeaturePyramid, LastLevelMaxPool
from .decoderhead import PredictHead

import torch.nn as nn
import torch
import torch.nn.functional as F
from functools import partial
from typing import List
import sys

sys.path.append('./modules')

from IMDLBenCo.registry import MODELS


@MODELS.register_module()
class IML_ViT(nn.Module):
    def __init__(
            self,
            # ViT backbone:
            input_size:int=1024,
            patch_size:int=16,
            embed_dim:int=768,
            vit_pretrain_path:str=None,  # wether to load pretrained weights
            # Simple_feature_pyramid_network:
            fpn_channels:int=256,
            fpn_scale_factors:List[int,]=(4.0, 2.0, 1.0, 0.5),
            # MLP embedding:
            mlp_embeding_dim:int=256,
            # Decoder head norm
            predict_head_norm:str="BN",
            # Edge loss:
            edge_lambda:int=20,
    ):
        """init iml_vit_model
        # TODO : add more args
        Args:
            input_size (int): size of the input image, defalut to 1024
            patch_size (int): patch size of Vision Transformer
            embed_dim (int): embedding dim for the ViT
            vit_pretrain_path (str): the path to initialize the model before start training
            fpn_channels (int): the number of embedding channels for simple feature pyraimd
            fpn_scale_factors(list(float, ...)) : the rescale factor for each SFPN layers.
            mlp_embedding dim: dim of mlp, i.e. decoder head
            predict_head_norm: the norm layer of predict head, need to select amoung 'BN', 'IN' and "LN"
                                We tested three different types of normalization in the decoder head, and they may yield different results due to dataset configurations and other factors.
                            Some intuitive conclusions are as follows:
                                - "LN" -> Layer norm : The fastest convergence, but poor generalization performance.
                                - "BN" Batch norm : When include authentic image during training, set batchsize = 2 may have poor performance. But if you can train with larger batchsize (e.g. A40 with 48GB memory can train with batchsize = 4) It may performs better.
                                - "IN" Instance norm : A form that can definitely converge, equivalent to a batchnorm with batchsize=1. When abnormal behavior is observed with BatchNorm, one can consider trying Instance Normalization. It's important to note that in this case, the settings should include setting track_running_stats and affine to True, rather than the default settings in PyTorch.
            edge_lambda(float) : the hyper-parameter for edge loss (lambda in our paper)
        """
        super(IML_ViT, self).__init__()
        self.input_size = input_size
        self.patch_size = patch_size
        # window attention vit
        self.encoder_net = window_attention_vit(
            img_size=input_size,
            patch_size=16,
            embed_dim=embed_dim,
            depth=12,
            num_heads=12,
            drop_path_rate=0.1,
            window_size=14,
            mlp_ratio=4,
            qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
            window_block_indexes=[
                # 2, 5, 8 11 for global attention
                0,
                1,
                3,
                4,
                6,
                7,
                9,
                10,
            ],
            residual_block_indexes=[],
            use_rel_pos=True,
            out_feature="last_feat",
        )
        self.vit_pretrain_path = vit_pretrain_path

        # simple feature pyramid network
        self.featurePyramid_net = SimpleFeaturePyramid(
            in_feature_shape=(1, embed_dim, 256, 256),
            out_channels=fpn_channels,
            scale_factors=fpn_scale_factors,
            top_block=LastLevelMaxPool(),
            norm="LN",
        )
        # MLP predict head
        self.predict_head = PredictHead(
            feature_channels=[fpn_channels for i in range(5)],
            embed_dim=mlp_embeding_dim,
            norm=predict_head_norm  # important! may influence the results
        )
        # Edge loss hyper-parameters    
        self.BCE_loss = nn.BCEWithLogitsLoss()
        self.edge_lambda = edge_lambda

        self.apply(self._init_weights)
        self._mae_init_weights()

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def _mae_init_weights(self):
        # Load MAE pretrained weights for Window Attention ViT encoder
        if self.vit_pretrain_path != None:
            self.encoder_net.load_state_dict(
                torch.load(self.vit_pretrain_path, map_location='cpu')['model'],  # BEIT MAE
                strict=False
            )
            print('load pretrained weights from \'{}\'.'.format(self.vit_pretrain_path))

    def forward(self, image: torch.Tensor, mask, edge_mask, shape=None, *args, **kwargs):
        x = self.encoder_net(image)
        x = self.featurePyramid_net(x)
        feature_list = []
        for k, v in x.items():
            feature_list.append(v)
        x = self.predict_head(feature_list)

        # up-sample to 1024x1024
        mask_pred = F.interpolate(x, size=(self.input_size, self.input_size), mode='bilinear', align_corners=False)

        # compute the loss
        predict_loss = self.BCE_loss(mask_pred, mask)

        edge_loss = F.binary_cross_entropy_with_logits(
            input=mask_pred,
            target=mask,
            weight=edge_mask
        ) * self.edge_lambda
        combined_loss = edge_loss + predict_loss

        mask_pred = torch.sigmoid(mask_pred)

        output_dict = {
            # loss for backward
            "backward_loss": combined_loss,
            # predicted mask, will calculate for metrics automatically
            "pred_mask": mask_pred,
            # predicted binaray label, will calculate for metrics automatically
            "pred_label": None,

            # ----values below is for visualization----
            # automatically visualize with the key-value pairs
            "visual_loss": {
                "predict_loss": predict_loss,
                "edge_loss": edge_loss,
                "combined_loss": combined_loss
            },

            "visual_image": {
                "pred_mask": mask_pred,
                "edge_mask": edge_mask
            }
            # -----------------------------------------
        }

        return output_dict



import sys
from pprint import pprint
sys.path.append(".")
from torch.utils.data import Dataset, DataLoader
import IMDLBenCo
import IMDLBenCo.datasets
from IMDLBenCo.datasets import ManiDataset
from IMDLBenCo.datasets.jpeg_dataset_deprecated import MetaCatnetDataset
from IMDLBenCo.registry import DATASETS
import torch

from IMDLBenCo.evaluation import generate_region_mask, cal_confusion_matrix, cal_F1
from IMDLBenCo.datasets import denormalize


data = IMDLBenCo.datasets.ManiDataset("/mnt/data0/public_datasets/IML/basic_eval_dataset", is_padding=True, edge_width= 7)
            # ['/mnt/data0/public_datasets/IML/IMD_20_1024', MetaCatnetDataset],
            # ['/mnt/data0/public_datasets/IML/tampCOCO/sp_COCO_list.json', MetaCatnetDataset],
            # ['/mnt/data0/public_datasets/IML/tampCOCO/cm_COCO_list.json', MetaCatnetDataset],
            # ['/mnt/data0/public_datasets/IML/tampCOCO/bcm_COCO_list.json', MetaCatnetDataset],
            # ['/mnt/data0/public_datasets/IML/tampCOCO/bcmc_COCO_list.json', MetaCatnetDataset]
# data = MetaCatnetDataset("/mnt/data0/public_datasets/IML/IMD_20_1024", is_padding=True, edge_width= 7)


batch_size = 3
dataloader = DataLoader(data, batch_size=batch_size, shuffle=True)

for i in dataloader:
    for key in i.keys():
        if isinstance(i[key], torch.Tensor) and key != 'shapes':
            print(key, i[key].shape)
        else:
            print(key,  i[key])
             
    import matplotlib.pyplot as plt
    plt.subplot(3, 2, 1)
    plt.imshow(i['mask'][0][0])
    plt.subplot(3, 2, 2)
    plt.imshow(i['mask'][1][0])
    
    # test for evaluation split
    # regin_mask = genertate_region_mask(i['masks'], i['shapes'])
    plt.subplot(3, 2, 3)
    plt.title('shape_mask')
    plt.imshow(i['shape_mask'][0][0])
    plt.subplot(3, 2, 4)
    plt.imshow(i['shape_mask'][1][0])
    
    plt.subplot(3, 2, 5)
    plt.imshow(i['image'][0][0])
    plt.subplot(3, 2, 6)
    plt.imshow(i['image'][1][0])
    
        
        
        
    plt.savefig("test.png")
    exit()






import cv2
import random
import numpy as np
# Augmentation library
import albumentations as albu
from albumentations.core.transforms_interface import DualTransform
from albumentations.pytorch import ToTensorV2

class RandomCopyMove(DualTransform):
    def __init__(self,
        max_h = 0.8,
        max_w = 0.8,
        min_h = 0.05,
        min_w = 0.05,
        mask_value = 255,
        always_apply = False,
        p = 0.5,  
    ):
        """Apply cope-move manipulation to the image, and change the respective region on the mask to <mask_value>

        Args:
            max_h (float, optional): (0~1), max window height rate to the full height of image . Defaults to 0.5.
            max_w (float, optional): (0~1), max window width rate to the full width of image . Defaults to 0.5.
            min_h (float, optional): (0~1), min window height rate to the full height of image . Defaults to 0.05.
            min_w (float, optional): (0~1), min window width rate to the full width of image . Defaults to 0.05.
            mask_value (int, optional): the value apply the tampered region on the mask. Defaults to 255.
            always_apply (bool, optional): _description_. Defaults to False.
            p (float, optional): _description_. Defaults to 0.5.
        """
        super(RandomCopyMove, self).__init__(always_apply, p)
        self.max_h = max_h
        self.max_w = max_w
        self.min_h = min_h
        self.min_w = min_w
        self.mask_value = mask_value
        
    def _get_random_window(
        self, 
        img_height, 
        img_width, 
        window_height = None, 
        window_width = None
    ):
        assert self.max_h < 1 and self.max_h > 0 
        assert self.max_w < 1 and self.max_w > 0
        assert self.min_w < 1 and self.min_w > 0
        assert self.min_h < 1 and self.min_h > 0
        
        l_min_h = int(img_height * self.min_h)
        l_min_w = int(img_width * self.min_w)
        l_max_h = int(img_height * self.max_h)
        l_max_w = int(img_width * self.max_w)
        
        if window_width == None or window_height == None:
            window_h = np.random.randint(l_min_h, l_max_h)
            window_w = np.random.randint(l_min_w, l_max_w)
        else:
            window_h = window_height
            window_w = window_width

        # position of left up corner of the window
        pos_h = np.random.randint(0, img_height - window_h)
        pos_w = np.random.randint(0, img_width - window_w)
        
        return pos_h, pos_w , window_h, window_w
        
        
    def apply(self, img: np.ndarray, **params) -> np.ndarray:
        image = img.copy()
        H, W, _ = image.shape
        # copy region:
        c_pos_h, c_pos_w, c_window_h, c_window_w = self._get_random_window(H, W)
        
        # past region, window size is defined by copy region:
        self.p_pos_h, self.p_pos_w, self.p_window_h, self.p_window_w = self._get_random_window(H, W, c_window_h, c_window_w)
        
        copy_region = image[
            c_pos_h: c_pos_h + c_window_h, 
            c_pos_w: c_pos_w + c_window_w, 
            : 
        ]
        image[
            self.p_pos_h : self.p_pos_h + self.p_window_h,
            self.p_pos_w : self.p_pos_w + self.p_window_w,
            :
        ] = copy_region
        return image
        

    def apply_to_mask(self, img: np.ndarray, **params) -> np.ndarray:
        """
        change the mask of manipulated region to 1
        """
    
        manipulated_region =  np.full((self.p_window_h, self.p_window_w), 1)
        img = img.copy()
        img[
            self.p_pos_h : self.p_pos_h + self.p_window_h,
            self.p_pos_w : self.p_pos_w + self.p_window_w,
        ] = self.mask_value
        return img
    # must be implemented for string output when print(RandomCopyMove())
    def get_transform_init_args_names(self):
        return ("max_h", "max_w", "min_h", "min_w", "mask_value", "always_apply", "p")
        
class RandomInpainting(DualTransform):
    def __init__(self,
        max_h = 0.8,
        max_w = 0.8,
        min_h = 0.05,
        min_w = 0.05,
        mask_value = 255,
        always_apply = False,
        p = 0.5,  
    ):
        super(RandomInpainting, self).__init__(always_apply, p)
        self.max_h = max_h
        self.max_w = max_w
        self.min_h = min_h
        self.min_w = min_w
        self.mask_value = mask_value
    def _get_random_window(
        self, 
        img_height, 
        img_width, 
    ):
        assert self.max_h < 1 and self.max_h > 0 
        assert self.max_w < 1 and self.max_w > 0
        assert self.min_w < 1 and self.min_w > 0
        assert self.min_h < 1 and self.min_h > 0
        
        l_min_h = int(img_height * self.min_h)
        l_min_w = int(img_width * self.min_w)
        l_max_h = int(img_height * self.max_h)
        l_max_w = int(img_width * self.max_w)
    
        window_h = np.random.randint(l_min_h, l_max_h)
        window_w = np.random.randint(l_min_w, l_max_w)

        # position of left up corner of the window
        pos_h = np.random.randint(0, img_height - window_h)
        pos_w = np.random.randint(0, img_width - window_w)
        
        return pos_h, pos_w , window_h, window_w
    
    def apply(self, img: np.ndarray, **params) -> np.ndarray:
        img = img.copy()
        img = np.uint8(img)
        H, W, C = img.shape
        mask = np.zeros((H, W), dtype=np.uint8)
        # inpainting region
        self.pos_h, self.pos_w , self.window_h, self.window_w = self._get_random_window(H, W)
        mask[
            self.pos_h : self.pos_h+ self.window_h,
            self.pos_w : self.pos_w + self.window_w,
        ] = 1
        inpaint_flag = cv2.INPAINT_TELEA if random.random() > 0.5 else cv2.INPAINT_NS
        img = cv2.inpaint(img, mask, 3,inpaint_flag)
        return img
    
    def apply_to_mask(self, img: np.ndarray, **params) -> np.ndarray:
        """
        change the mask of manipulated region to 1
        """
        img = img.copy()
        img[
            self.pos_h : self.pos_h+ self.window_h,
            self.pos_w : self.pos_w + self.window_w,
        ] = self.mask_value
        return img
    
    # must be implemented for string output when print(RandomInpainting())
    def get_transform_init_args_names(self):
        return ("max_h", "max_w", "min_h", "min_w", "mask_value", "always_apply", "p")

def get_albu_transforms(type_ = 'train', output_size = (1024, 1024)):
    """get albumentations transforms
        
        type_ (str): 
            if 'train', then return train transforms with
                random scale, flip, rotate, brightness, contrast, and GaussianBlur augmentation.
            if 'test' then return test transforms 
            if 'pad' then return zero-padding transforms
    """
    
    assert type_ in ['train', 'test', 'pad', 'resize'] , "type_ must be 'train' or 'test' of 'pad' "
    trans = None
    if type_ == 'train':
        trans = albu.Compose([
            # Rescale the input image by a random factor between 0.8 and 1.2
            albu.RandomScale(scale_limit=0.2, p=1), 
            RandomCopyMove(p = 0.1),
            RandomInpainting(p = 0.1),
            # Flips
            # albu.Resize(512, 512),
            albu.HorizontalFlip(p=0.5),
            albu.VerticalFlip(p=0.5),
            # Brightness and contrast fluctuation
            albu.RandomBrightnessContrast(
                brightness_limit=(-0.1, 0.1),
                contrast_limit=0.1,
                p=1
            ),
            albu.ImageCompression(
                quality_lower = 70,
                quality_upper = 100,
                p = 0.2
            ),
            # Rotate
            albu.RandomRotate90(p=0.5),
            # Blur
            albu.GaussianBlur(
                blur_limit = (3, 7),
                p = 0.2
            ),
        ])
    
    if type_ == 'test':
        trans = None
        trans = albu.Compose([
        # ---Blow for robustness evalution---
        # albu.Resize(512, 512),
        #   albu.JpegCompression(
        #         quality_lower = 89,
        #         quality_upper = 90,
        #         p = 1
        #   ),
        #  albu.GaussianBlur(
        #         blur_limit = (5, 5),
        #         p = 1
        #     ),
        
        # albu.GaussNoise(
        #     var_limit=(15, 15),
        #     p = 1
        # )
        ])
        
    if type_ == 'pad':
        trans = albu.Compose([
            albu.PadIfNeeded(          
                min_height=output_size[0],
                min_width=output_size[1], 
                border_mode=0, 
                value=0, 
                position= 'top_left',
                mask_value=0),
            albu.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            albu.Crop(0, 0, output_size[0], output_size[1]),
            ToTensorV2(transpose_mask=True)
        ])
    if type_ == 'resize':
        trans = albu.Compose([
            albu.Resize(output_size[0], output_size[1]),
            albu.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            albu.Crop(0, 0, output_size[0], output_size[1]),
            ToTensorV2(transpose_mask=True)
        ])

    return trans


import torch
import torch.nn as nn
import torch.nn.functional as F
from .seg_hrnet_config import get_hrnet_cfg


class NonLocalMask(nn.Module):
    def __init__(self, in_channels, reduce_scale):
        super(NonLocalMask, self).__init__()

        self.r = reduce_scale

        # input channel number
        self.ic = in_channels * self.r * self.r

        # middle channel number
        self.mc = self.ic

        self.g = nn.Conv2d(in_channels=self.ic, out_channels=self.ic,
                           kernel_size=1, stride=1, padding=0)

        self.theta = nn.Conv2d(in_channels=self.ic, out_channels=self.mc,
                               kernel_size=1, stride=1, padding=0)
        self.phi = nn.Conv2d(in_channels=self.ic, out_channels=self.mc,
                             kernel_size=1, stride=1, padding=0)

        self.W_s = nn.Conv2d(in_channels=in_channels, out_channels=in_channels,
                             kernel_size=1, stride=1, padding=0)

        self.W_c = nn.Conv2d(in_channels=in_channels, out_channels=in_channels,
                             kernel_size=1, stride=1, padding=0)

        self.gamma_s = nn.Parameter(torch.ones(1))

        self.gamma_c = nn.Parameter(torch.ones(1))

        self.getmask = nn.Sequential(
            nn.Conv2d(in_channels=in_channels, out_channels=16, kernel_size=3, stride=1, padding=1),
            nn.ReLU(),
            nn.Conv2d(in_channels=16, out_channels=1, kernel_size=3, stride=1, padding=1))

    def forward(self, x):
        """
            inputs :
                x : input feature maps( B X C X H X W)
            value :
                f: B X (HxW) X (HxW)
                ic: intermediate channels
                z: feature maps( B X C X H X W)
            output:
                mask: feature maps( B X 1 X H X W)
        """

        b, c, h, w = x.shape

        x1 = x.reshape(b, self.ic, h // self.r, w // self.r)

        # g x
        g_x = self.g(x1).view(b, self.ic, -1)
        g_x = g_x.permute(0, 2, 1)

        # theta
        theta_x = self.theta(x1).view(b, self.mc, -1)

        theta_x_s = theta_x.permute(0, 2, 1)
        theta_x_c = theta_x

        # phi x
        phi_x = self.phi(x1).view(b, self.mc, -1)

        phi_x_s = phi_x
        phi_x_c = phi_x.permute(0, 2, 1)

        # non-local attention
        f_s = torch.matmul(theta_x_s, phi_x_s)
        f_s_div = F.softmax(f_s, dim=-1)

        f_c = torch.matmul(theta_x_c, phi_x_c)
        f_c_div = F.softmax(f_c, dim=-1)

        # get y_s
        y_s = torch.matmul(f_s_div, g_x)
        y_s = y_s.permute(0, 2, 1).contiguous()
        y_s = y_s.view(b, c, h, w)

        # get y_c
        y_c = torch.matmul(g_x, f_c_div)
        y_c = y_c.view(b, c, h, w)

        # get z
        z = x + self.gamma_s * self.W_s(y_s) + self.gamma_c * self.W_c(y_c)

        # get mask
        # mask = torch.sigmoid(self.getmask(z.clone()))
        mask = self.getmask(z.clone())      # No sigmoid since AMP (自动半精度)
        return mask, z


class NLCDetection(nn.Module):
    def __init__(self, args):
        super(NLCDetection, self).__init__()

        self.crop_size = args['crop_size']

        FENet_cfg = get_hrnet_cfg()

        num_channels = FENet_cfg['STAGE4']['NUM_CHANNELS']

        feat1_num, feat2_num, feat3_num, feat4_num = num_channels

        self.getmask4 = NonLocalMask(feat4_num, 1)
        self.getmask3 = NonLocalMask(feat3_num, 2)
        self.getmask2 = NonLocalMask(feat2_num, 2)
        self.getmask1 = NonLocalMask(feat1_num, 4)

    def forward(self, feat):
        """
            inputs :
                feat : a list contains features from s1, s2, s3, s4
            output:
                mask1: output mask ( B X 1 X H X W)
                pred_cls: output cls (B X 4)
        """
        s1, s2, s3, s4 = feat

        if s1.shape[2:] == self.crop_size:
            pass
        else:
            s1 = F.interpolate(s1, size=self.crop_size, mode='bilinear', align_corners=True)
            s2 = F.interpolate(s2, size=[i // 2 for i in self.crop_size], mode='bilinear', align_corners=True)
            s3 = F.interpolate(s3, size=[i // 4 for i in self.crop_size], mode='bilinear', align_corners=True)
            s4 = F.interpolate(s4, size=[i // 8 for i in self.crop_size], mode='bilinear', align_corners=True)

        mask4, z4 = self.getmask4(s4)
        mask4U = F.interpolate(mask4, size=s3.size()[2:], mode='bilinear', align_corners=True)

        s3 = s3 * mask4U
        mask3, z3 = self.getmask3(s3)
        mask3U = F.interpolate(mask3, size=s2.size()[2:], mode='bilinear', align_corners=True)

        s2 = s2 * mask3U
        mask2, z2 = self.getmask2(s2)
        mask2U = F.interpolate(mask2, size=s1.size()[2:], mode='bilinear', align_corners=True)

        s1 = s1 * mask2U
        mask1, z1 = self.getmask1(s1)

        return mask1, mask2, mask3, mask4


import timm
import torch 
import torch
import torch.nn as nn
import timm
import torch.nn.functional as F
import sys
import os
sys.path.append('.')

from timm.models.layers import DropPath, to_2tuple, trunc_normal_
from .extractor.high_frequency_feature_extraction import HighDctFrequencyExtractor
from .extractor.low_frequency_feature_extraction import LowDctFrequencyExtractor
import math
from functools import partial
from IMDLBenCo.registry import MODELS

class ConvNeXt(timm.models.convnext.ConvNeXt):
    def __init__(self,conv_pretrain=False):
        super(ConvNeXt, self).__init__(depths=(3, 3, 9, 3), dims=(96, 192, 384, 768))
        if conv_pretrain:
            print("Load Convnext pretrain.")
            model = timm.create_model('convnext_tiny', pretrained=True)
            self.load_state_dict(model.state_dict())
        original_first_layer = self.stem[0]
        new_first_layer = nn.Conv2d(6, original_first_layer.out_channels,
                                        kernel_size=original_first_layer.kernel_size, stride=original_first_layer.stride,
                                        padding=original_first_layer.padding, bias=False)
        new_first_layer.weight.data[:, :3, :, :] = original_first_layer.weight.data.clone()[:, :3, :, :]
        new_first_layer.weight.data[:, 3:, :, :] = torch.nn.init.kaiming_normal_(new_first_layer.weight[:, 3:, :, :])
        self.stem[0] = new_first_layer

    def forward_features(self, x):
        x = self.stem(x)
        out = []
        for stage in self.stages:
            x = stage(x)
            out.append(x)
        x = self.norm_pre(x)
        return x , out
    def forward(self, image, mask=None, *args, **kwargs):

        feature,out = self.forward_features(image)

        return feature,out

class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = self.fc1(x)
        x = self.dwconv(x, H, W)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0., sr_ratio=1):
        super().__init__()
        assert dim % num_heads == 0, f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        B, N, C = x.shape
        q = self.q(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)

        if self.sr_ratio > 1:
            x_ = x.permute(0, 2, 1).reshape(B, C, H, W)
            x_ = self.sr(x_).reshape(B, C, -1).permute(0, 2, 1)
            x_ = self.norm(x_)
            kv = self.kv(x_).reshape(B, -1, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        else:
            kv = self.kv(x).reshape(B, -1, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.float()
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class Block(nn.Module):

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, sr_ratio=1):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale,
            attn_drop=attn_drop, proj_drop=drop, sr_ratio=sr_ratio)
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))

        return x


class OverlapPatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """

    def __init__(self, img_size=224, patch_size=7, stride=4, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)

        self.img_size = img_size
        self.patch_size = patch_size
        self.H, self.W = img_size[0] // patch_size[0], img_size[1] // patch_size[1]
        self.num_patches = self.H * self.W
        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=stride,
                              padding=(patch_size[0] // 2, patch_size[1] // 2))
        self.norm = nn.LayerNorm(embed_dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x):
        x = self.proj(x)
        _, _, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return x, H, W



class UpsampleConcatConvSegformer(nn.Module):
    def __init__(self):
        super(UpsampleConcatConvSegformer, self).__init__()
        # 192到96的上采样，单次上采样
        self.upsample1 = nn.ConvTranspose2d(128, 64, kernel_size=4, stride=2, padding=1)

        # 384到96的上采样，两次上采样，逐步降低通道数
        self.upsample2 = nn.Sequential(
            nn.ConvTranspose2d(320, 128, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(128, 64, kernel_size=4, stride=2, padding=1)
        )

        # 768到96的上采样，三次上采样，逐步降低通道数
        self.upsample3 = nn.Sequential(
            nn.ConvTranspose2d(512, 320, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(320, 128, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(128, 64, kernel_size=4, stride=2, padding=1)
        )


    def forward(self, inputs):
        # 上采样
        x1,x2,x3,x4 = inputs
        up2 = self.upsample1(x2)
        up3 = self.upsample2(x3)
        up4 = self.upsample3(x4)
        
        x = torch.cat([x1, up2, up3, up4], dim=1)
        return x



# class mit_b3(MixVisionTransformer):
#     def __init__(self, **kwargs):
#         super(mit_b3, self).__init__(
#             patch_size=4, embed_dims=[64, 128, 320, 512], num_heads=[1, 2, 5, 8], mlp_ratios=[4, 4, 4, 4],
#             qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), depths=[3, 4, 18, 3], sr_ratios=[8, 4, 2, 1],
#             drop_rate=0.0, drop_path_rate=0.1)
class MixVisionTransformer(nn.Module):
    def __init__(self,pretrain_path=None, img_size=512, patch_size=4, in_chans=3,embed_dims=[64, 128, 320, 512],num_heads=[1, 2, 5, 8], mlp_ratios=[4, 4, 4, 4], qkv_bias=True, qk_scale=None, drop_rate=0.0,
                 attn_drop_rate=0., drop_path_rate=0.1, norm_layer=partial(nn.LayerNorm, eps=1e-6),
                 depths=[3, 4, 18, 3], sr_ratios=[8, 4, 2, 1]):
        super().__init__()
        self.depths = depths

        # patch_embed
        self.patch_embed1 = OverlapPatchEmbed(img_size=img_size, patch_size=7, stride=4, in_chans=in_chans,
                                              embed_dim=embed_dims[0])

        self.patch_embed2 = OverlapPatchEmbed(img_size=img_size // 4, patch_size=3, stride=2, in_chans=embed_dims[0],
                                              embed_dim=embed_dims[1])
        self.patch_embed3 = OverlapPatchEmbed(img_size=img_size // 8, patch_size=3, stride=2, in_chans=embed_dims[1],
                                              embed_dim=embed_dims[2])
        self.patch_embed4 = OverlapPatchEmbed(img_size=img_size // 16, patch_size=3, stride=2, in_chans=embed_dims[2],
                                              embed_dim=embed_dims[3])

        # transformer encoder
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule
        cur = 0
        self.block1 = nn.ModuleList([Block(
            dim=embed_dims[0], num_heads=num_heads[0], mlp_ratio=mlp_ratios[0], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[0])
            for i in range(depths[0])])
        self.norm1 = norm_layer(embed_dims[0])

        cur += depths[0]
        self.block2 = nn.ModuleList([Block(
            dim=embed_dims[1], num_heads=num_heads[1], mlp_ratio=mlp_ratios[1], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[1])
            for i in range(depths[1])])
        self.norm2 = norm_layer(embed_dims[1])

        cur += depths[1]
        self.block3 = nn.ModuleList([Block(
            dim=embed_dims[2], num_heads=num_heads[2], mlp_ratio=mlp_ratios[2], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[2])
            for i in range(depths[2])])
        self.norm3 = norm_layer(embed_dims[2])

        cur += depths[2]
        self.block4 = nn.ModuleList([Block(
            dim=embed_dims[3], num_heads=num_heads[3], mlp_ratio=mlp_ratios[3], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer,
            sr_ratio=sr_ratios[3])
            for i in range(depths[3])])
        self.norm4 = norm_layer(embed_dims[3])
        if pretrain_path is not None:
            print("Load segformer pretrain pth.")
            self.load_state_dict(torch.load(pretrain_path),
                                strict=False)
        original_first_layer = self.patch_embed1.proj
        new_first_layer = nn.Conv2d(6, original_first_layer.out_channels,
                                        kernel_size=original_first_layer.kernel_size, stride=original_first_layer.stride,
                                        padding=original_first_layer.padding, bias=False)
        new_first_layer.weight.data[:, :3, :, :] = original_first_layer.weight.data.clone()[:, :3, :, :]
    
        new_first_layer.weight.data[:, 3:, :, :] = torch.nn.init.kaiming_normal_(new_first_layer.weight[:, 3:, :, :])
        self.patch_embed1.proj = new_first_layer


    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward_features(self, x):
        B = x.shape[0]
        outs = []

        # stage 1
        x, H, W = self.patch_embed1(x)
        for i, blk in enumerate(self.block1):
            x = blk(x, H, W)
        x = self.norm1(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 2
        x, H, W = self.patch_embed2(x)
        for i, blk in enumerate(self.block2):
            x = blk(x, H, W)
        x = self.norm2(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 3
        x, H, W = self.patch_embed3(x)
        for i, blk in enumerate(self.block3):
            x = blk(x, H, W)
        x = self.norm3(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 4
        x, H, W = self.patch_embed4(x)
        for i, blk in enumerate(self.block4):
            x = blk(x, H, W)
        x = self.norm4(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)
        return x,outs

    def forward(self, x):
        x,outs = self.forward_features(x)
        return x,outs 


class DWConv(nn.Module):
    def __init__(self, dim=768):
        super(DWConv, self).__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, bias=True, groups=dim)

    def forward(self, x, H, W):
        B, N, C = x.shape
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)
        x = x.flatten(2).transpose(1, 2)

        return x
import torch
import torch.nn as nn
import torch.nn.functional as F

class UpsampleConcatConv(nn.Module):
    def __init__(self):
        super(UpsampleConcatConv, self).__init__()
        # 192到96的上采样，单次上采样
        self.upsamplec2 = nn.ConvTranspose2d(192, 96, kernel_size=4, stride=2, padding=1)
        self.upsamples2 = nn.ConvTranspose2d(128, 64, kernel_size=4, stride=2, padding=1)
        # 384到96的上采样，两次上采样，逐步降低通道数
        self.upsamplec3 = nn.Sequential(
            nn.ConvTranspose2d(384, 192, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(192, 96, kernel_size=4, stride=2, padding=1)
        )
        self.upsamples3 = nn.Sequential(
            nn.ConvTranspose2d(320, 128, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(128, 64, kernel_size=4, stride=2, padding=1)
        )

        # 768到96的上采样，三次上采样，逐步降低通道数
        self.upsamplec4 = nn.Sequential(
            nn.ConvTranspose2d(768, 384, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(384, 192, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(192, 96, kernel_size=4, stride=2, padding=1)
        )
        self.upsamples4 = nn.Sequential(
            nn.ConvTranspose2d(512, 320, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(320, 128, kernel_size=4, stride=2, padding=1),
            nn.ConvTranspose2d(128, 64, kernel_size=4, stride=2, padding=1)
        )


    def forward(self, inputs):
        # 上采样
        c1,c2,c3,c4,s1,s2,s3,s4 = inputs

        c2 = self.upsamplec2(c2)
        c3 = self.upsamplec3(c3)
        c4 = self.upsamplec4(c4)
        s2 = self.upsamples2(s2)
        s3 = self.upsamples3(s3)
        s4 = self.upsamples4(s4)
        
        # 拼接四个tensor
        x = torch.cat([c1,c2,c3,c4,s1,s2,s3,s4 ], dim=1)
        features = [c1,c2,c3,c4,s1,s2,s3,s4]
        # shortcut = x
        # x = x.permute(0, 2, 3, 1)
        # x = self.fc2(self.act(self.fc1(x)))
        # x = x.permute(0, 3, 1, 2)
        # x = x + shortcut
        # 1x1卷积
        return x, features

class LayerNorm2d(nn.LayerNorm):
    """ LayerNorm for channels of '2D' spatial NCHW tensors """
    def __init__(self, num_channels, eps=1e-6, affine=True):
        super().__init__(num_channels, eps=eps, elementwise_affine=affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.permute(0, 2, 3, 1)
        x = F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        x = x.permute(0, 3, 1, 2)
        return x
    
class ScoreNetwork(nn.Module):
    def __init__(self):
        super(ScoreNetwork, self).__init__()
        self.conv1 = nn.Conv2d(9, 192, kernel_size=7, stride=2, padding=3)
        self.invert = nn.Sequential(LayerNorm2d(192),
                                    nn.Conv2d(192, 192, kernel_size=3, stride=1, padding=1),
                                    nn.Conv2d(192, 768, kernel_size=1),
                                    nn.Conv2d(768, 192, kernel_size=1),
                                    nn.GELU())
        self.conv2 = nn.Conv2d(192, 8,  kernel_size=7, stride=2, padding=3)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        x = self.conv1(x)
        short_cut = x
        x = self.invert(x)
        x = short_cut + x
        x = self.conv2(x)
        x = x.float()
        x = self.softmax(x)
        return x


@MODELS.register_module()
class Mesorch(nn.Module):
    def __init__(self, seg_pretrain_path=None, conv_pretrain=False, if_predict_label=False):
        super(Mesorch, self).__init__()
        self.convnext = ConvNeXt(conv_pretrain)
        self.segformer = MixVisionTransformer(seg_pretrain_path)
        self.upsample = UpsampleConcatConv()
        self.low_dct = LowDctFrequencyExtractor()
        self.high_dct = HighDctFrequencyExtractor()
        # 使用1x1的卷积将4个192通道合并为1通道
        self.inverse = nn.ModuleList([nn.Conv2d(96, 1, 1) for _ in range(4)]+[nn.Conv2d(64, 1, 1) for _ in range(4)])
        self.gate = ScoreNetwork()
        # 最后调整到512x512大小
        self.resize = nn.Upsample(size=(512, 512), mode='bilinear', align_corners=True)
        self.loss_fn = nn.BCEWithLogitsLoss()
        # self.if_predict_label = if_predict_label
        # if if_predict_label:
        #     self.pooled = nn.AdaptiveAvgPool2d(1)
        #     self.label_head = nn.Sequential(
        #         nn.Linear(2464, 256),
        #         nn.ReLU(),
        #         nn.Linear(256, 1)
        #     )
    def forward_feature(self, image, *args, **kwargs):
        high_freq = self.high_dct(image)
        low_freq = self.low_dct(image)
        input_high = torch.concat([image,high_freq],dim=1)
        input_low = torch.concat([image,low_freq],dim=1)
        input_all = torch.concat([image,high_freq,low_freq],dim=1)
        _,outs1 = self.convnext(input_high)
        _,outs2 = self.segformer(input_low)
        gate_outputs = self.gate(input_all)
        features = outs1 + outs2
        
        return features, gate_outputs
    
    def forward(self, image, mask, label, *args, **kwargs):
        features_raw, gate_outputs = self.forward_feature(image)
        x, features = self.upsample(features_raw)
        reduced = torch.cat([self.inverse[i](features[i]) for i in range(8)], dim=1)
        pred_mask = torch.sum(gate_outputs * reduced, dim=1,keepdim=True)
        pred_label = None
        label_loss = torch.tensor(0.0)
        # if self.if_predict_label:
            # weighted_features = []
            # pooled_features = [self.pooled(f).squeeze(-1).squeeze(-1) for f in features_raw]  # List of [B, C]
            # pooled_gate = self.pooled(gate_outputs)
            # for i in range(8):
            #     b = pooled_gate.shape[0]
            #     w = pooled_gate[:, i].view(b, 1)  # reshape 成 (b, 1)
            #     f = pooled_features[i]  # 取出第 i 个特征
            #     weighted_f = f * w  # 自动广播乘法
            #     weighted_features.append(weighted_f)
            # pooled_features = torch.cat(weighted_features, dim=1)  # [B, total_dim]
            # pred_label = self.label_head(pooled_features).squeeze(-1)
            # label_loss = self.loss_fn(pred_label,label.float())
            # pred_label = torch.sigmoid(pred_label)
        # 调整大小到512x512
        pred_mask = self.resize(pred_mask)
        mask_loss = self.loss_fn(pred_mask,mask)
        loss = mask_loss + label_loss
        pred_mask = pred_mask.float()
        pred_mask = torch.sigmoid(pred_mask)
        
        output_dict = {
            # loss for backward
            "backward_loss": loss,
            # predicted mask, will calculate for metrics automatically
            "pred_mask": pred_mask,
            # predicted binaray label, will calculate for metrics automatically
            "pred_label": pred_label,

            # ----values below is for visualization----
            # automatically visualize with the key-value pairs
            "visual_loss": {
                "predict_loss": loss,
                'predict_maks_loss': mask_loss,
                # 'predcit_label_loss': label_loss
            },

            "visual_image": {
                "pred_mask": pred_mask,
            }
            # -----------------------------------------
        }
        return output_dict

if __name__ == "__main__":
    print(MODELS)


import sys
from pprint import pprint
sys.path.append(".")

import IMDLBenCo
from IMDLBenCo.evaluation import PixelF1, ImageF1

list_e = [
    PixelF1(),
    ImageF1()
]


print(list_e[0], list_e[1])



def print_fun(obj):
    for i in obj:
        print(i)
        
        
        
print_fun(list_e)


import torch

weight_path = 'MantraNetv4.pt'
state_dict = torch.load(weight_path)

new_state_dict =  {'model': {}}
for key, value in state_dict.items():
    new_state_dict['model'][key] = value

new_weight_path = '../../../eval_dir/checkpoint-0.pth'
torch.save(new_state_dict, new_weight_path)

print(f"Modified weights saved to {new_weight_path}")



import numpy as np

from .base_cam_hack import BaseCAM


class GradCAM(BaseCAM):
    def __init__(self, model, target_layers,
                 reshape_transform=None):
        super(
            GradCAM,
            self).__init__(
            model,
            target_layers,
            reshape_transform)

    def get_cam_weights(self,
                        input_tensor,
                        target_layer,
                        target_category,
                        activations,
                        grads):
        # 2D image
        if len(grads.shape) == 4:
            return np.mean(grads, axis=(2, 3))
        
        # 3D image
        elif len(grads.shape) == 5:
            return np.mean(grads, axis=(2, 3, 4))
        
        else:
            raise ValueError("Invalid grads shape." 
                             "Shape of grads should be 4 (2D image) or 5 (3D image).")


import numpy as np
from IMDLBenCo.datasets.utils import read_jpeg_from_memory
from IMDLBenCo.registry import POSTFUNCS

@POSTFUNCS.register_module()
def cat_net_post_func(data_dict):
    tp_img = data_dict['image']
    DCT_coef, qtables = __get_jpeg_info(tp_img)
    data_dict['DCT_coef'] = DCT_coef[0]
    data_dict['qtables'] = qtables[0]
    
def __get_jpeg_info(image_tensor):
    """
    :param im_path: JPEG image path
    :return: DCT_coef (Y,Cb,Cr), qtables (Y,Cb,Cr)
    """
    num_channels = 1
    jpeg = read_jpeg_from_memory(image_tensor)

    # determine which axes to up-sample
    ci = jpeg.comp_info
    need_scale = [[ci[i].v_samp_factor, ci[i].h_samp_factor] for i in range(num_channels)]
    if num_channels == 3:
        if ci[0].v_samp_factor == ci[1].v_samp_factor == ci[2].v_samp_factor:
            need_scale[0][0] = need_scale[1][0] = need_scale[2][0] = 2
        if ci[0].h_samp_factor == ci[1].h_samp_factor == ci[2].h_samp_factor:
            need_scale[0][1] = need_scale[1][1] = need_scale[2][1] = 2
    else:
        need_scale[0][0] = 2
        need_scale[0][1] = 2

    # up-sample DCT coefficients to match image size
    DCT_coef = []
    for i in range(num_channels):
        r, c = jpeg.coef_arrays[i].shape
        coef_view = jpeg.coef_arrays[i].reshape(r//8, 8, c//8, 8).transpose(0, 2, 1, 3)
        # case 1: row scale (O) and col scale (O)
        if need_scale[i][0]==1 and need_scale[i][1]==1:
            out_arr = np.zeros((r * 2, c * 2))
            out_view = out_arr.reshape(r * 2 // 8, 8, c * 2 // 8, 8).transpose(0, 2, 1, 3)
            out_view[::2, ::2, :, :] = coef_view[:, :, :, :]
            out_view[1::2, ::2, :, :] = coef_view[:, :, :, :]
            out_view[::2, 1::2, :, :] = coef_view[:, :, :, :]
            out_view[1::2, 1::2, :, :] = coef_view[:, :, :, :]

        # case 2: row scale (O) and col scale (X)
        elif need_scale[i][0]==1 and need_scale[i][1]==2:
            out_arr = np.zeros((r * 2, c))
            DCT_coef.append(out_arr)
            out_view = out_arr.reshape(r*2//8, 8, c // 8, 8).transpose(0, 2, 1, 3)
            out_view[::2, :, :, :] = coef_view[:, :, :, :]
            out_view[1::2, :, :, :] = coef_view[:, :, :, :]

        # case 3: row scale (X) and col scale (O)
        elif need_scale[i][0]==2 and need_scale[i][1]==1:
            out_arr = np.zeros((r, c * 2))
            out_view = out_arr.reshape(r // 8, 8, c * 2 // 8, 8).transpose(0, 2, 1, 3)
            out_view[:, ::2, :, :] = coef_view[:, :, :, :]
            out_view[:, 1::2, :, :] = coef_view[:, :, :, :]

        # case 4: row scale (X) and col scale (X)
        elif need_scale[i][0]==2 and need_scale[i][1]==2:
            out_arr = np.zeros((r, c))
            out_view = out_arr.reshape(r // 8, 8, c // 8, 8).transpose(0, 2, 1, 3)
            out_view[:, :, :, :] = coef_view[:, :, :, :]

        else:
            raise KeyError("Something wrong here.")

        DCT_coef.append(out_arr)

    # quantization tables
    qtables = [jpeg.quant_tables[ci[i].quant_tbl_no].astype(np.float64) for i in range(num_channels)]

    return DCT_coef, qtables



import torch
import torch.nn as nn
import timm
import torch.nn.functional as F
from IMDLBenCo.registry import MODELS

from ..extractors.high_frequency_feature_extraction import FFTExtractor, DCTExtractor
from ..extractors.sobel import SobelFilter
from ..extractors.bayar_conv import BayerConv
from ..extractors.srm_filter import SRMConv2D

class ViT(timm.models.vision_transformer.VisionTransformer):
    def __init__(self, input_head=None, num_channels=3,pretrained=True):
        super(ViT, self).__init__()
        if pretrained:
            model = timm.create_model('vit_base_patch16_224', pretrained=True)
            self.load_state_dict(model.state_dict())
        original_first_layer = self.patch_embed.proj
        self.patch_embed.strict_img_size=False
        if input_head != None:
            self.input_head = input_head
            new_first_layer = nn.Conv2d(num_channels + 3, original_first_layer.out_channels,
                                        kernel_size=original_first_layer.kernel_size, stride=original_first_layer.stride,
                                        padding=original_first_layer.padding, bias=False)
            new_first_layer.weight.data[:, :3, :, :] = original_first_layer.weight.data.clone()[:, :3, :, :]
            if num_channels > 0:
                new_first_layer.weight.data[:, 3:, :, :] = torch.nn.init.kaiming_normal_(new_first_layer.weight[:, 3:, :, :])
            self.patch_embed.proj = new_first_layer
        else:
            self.input_head = None
        new_length = 1024
        self.deconv_model = nn.Sequential(
            nn.ConvTranspose2d(768, 256, kernel_size=3, stride=2, padding=1, output_padding=1),  # 输出: 1, 256, 64, 64
            nn.ReLU(),
            nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, output_padding=1),  # 输出: 1, 128, 128, 128
            nn.ReLU(),
            nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, output_padding=1),   # 输出: 1, 64, 256, 256
            nn.ReLU(),
            nn.ConvTranspose2d(64, 1, kernel_size=3, stride=2, padding=1, output_padding=1)      # 输出: 1, 1, 512, 512
        )
        
        original_pos_embed = model.pos_embed[:, 1:, :]
        new_pos_embed = F.interpolate(original_pos_embed.permute(0, 2, 1), size = new_length, mode='nearest').permute(0, 2, 1)

        # Create new positional embeddings tensor with space for class token
        new_pos_embed_with_cls = torch.zeros(1,new_length+1, 768, device=original_pos_embed.device)
        new_pos_embed_with_cls[:, 1:, :] = new_pos_embed  # fill in the interpolated embeddings
        new_pos_embed_with_cls[:, 0, :] = model.pos_embed[:, 0, :]  # copy the class token embeddings
        # 
        # Replace the original positional embeddings with the new ones
        self.pos_embed = nn.Parameter(new_pos_embed_with_cls)
        self.loss_fun = nn.BCEWithLogitsLoss()


    def forward(self, image, mask, *args, **kwargs):
        if self.input_head != None:
            feature = self.input_head(image)
            input = torch.concat([image,feature],dim=1)
        else:
            input = image
        x = self.patch_embed(input)
        x = self._pos_embed(x)
        x = self.blocks(x)
        feature = self.norm(x)
        h = w = int((feature.shape[1])**0.5)
        feature = feature.permute(0,2,1)[:,:,1:1025].reshape(image.shape[0],768,h,w)
        mask_pred = self.deconv_model(feature)
        loss = self.loss_fun(mask_pred,mask)
        mask_pred = torch.sigmoid(mask_pred)
        output_dict = {
            # loss for backward
            "backward_loss": loss,
            # predicted mask, will calculate for metrics automatically
            "pred_mask": mask_pred,
            # predicted binaray label, will calculate for metrics automatically
            "pred_label": None,

            # ----values below is for visualization----
            # automatically visualize with the key-value pairs
            "visual_loss": {
                "predict_loss": loss,
            },

            "visual_image": {
                "pred_mask": mask_pred,
            }
            # -----------------------------------------
        }
        return output_dict

@MODELS.register_module()
def vit16():
    return ViT(None)

@MODELS.register_module()
def fft_vit16():
    return ViT(FFTExtractor(), 3)

@MODELS.register_module()
def dct_vit16():
    return ViT(DCTExtractor(), 3)

@MODELS.register_module()
def sobel_vit16():
    return ViT(SobelFilter(), 1)

@MODELS.register_module()
def bayar_vit16():
    return ViT(BayerConv(), 3)

@MODELS.register_module()
def srm_vit16():
    return ViT(SRMConv2D(), 9)

if __name__ == '__main__':
    from fvcore.nn import FlopCountAnalysis, parameter_count_table,flop_count_table

    model = ViT()
    flops = FlopCountAnalysis(model,(torch.randn(1,3,512,512),torch.randn(1,1,512,512)))
    # # 打印模型的参数信息
    print(flop_count_table(flops))
